"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.nd_hist_generator.outputs import *
from os import sep


class TestOutput(TestCase):
    """ Test how well we parse the outputs """
    parser_objects = []
    parser_objects.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_1.hdf5",
                                       flim=True))
    parser_objects.append(OutputParser(output_dict={'memory': True}, filename="tests_data" + sep + f"output_2.hdf5",
                                       reprate=160.3e6, flim=True, z_pixels=10, num_of_frames=2))
    parser_objects.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_1.hdf5",
                                       num_of_frames=2))

    def test_bins_bet_pulses(self):
        bins_bet_pulses = [16, 8, 1]
        for obj, bins in zip(self.parser_objects, bins_bet_pulses):
            self.assertEqual(obj.bins_bet_pulses, bins)


    def test_determine_shape(self):
        shapes = [(1, 512, 512, 16), (2, 512, 512, 10, 8), (2, 512, 512)]
        for shape, obj in zip(shapes, self.parser_objects):
            self.assertEqual(shape, obj.determine_data_shape_full())

    def test_non_squeezed_shapes(self):
        shapes = [(1, 512, 512, 16, 1), (1, 1, 512, 100), (10, 512, 1, 16), (10, 512, 1)]
        squeezed_shapes = [(1, 512, 512, 16), (1, 512, 100), (10, 512, 16), (10, 512)]
        output_obj = []
        output_obj.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_3.hdf5",
                                       z_pixels=16, num_of_frames=1))
        output_obj.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_4.hdf5",
                                       z_pixels=100, num_of_frames=1, x_pixels=1))
        output_obj.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_5.hdf5",
                                       z_pixels=16, num_of_frames=10, y_pixels=1))
        output_obj.append(OutputParser(output_dict={}, filename="tests_data" + sep + f"output_6.hdf5",
                                       num_of_frames=10, y_pixels=1))
        for shape, obj in zip(squeezed_shapes, output_obj):
            self.assertEqual(shape, obj.determine_data_shape_full())



import unittest
import numpy as np

from pysight.ascii_list_file_parser.tabulation import *


class TestTabulation(unittest.TestCase):

    tab = Tabulate(dict_of_inputs={}, data=np.array([]),
                   dict_of_slices_hex={})
    def test_conversion_hex_to_bits(self):
        diction = \
            {
                '0': '0000',
                '1': '0001',
                '2': '0010',
                '3': '0011',
                '4': '0100',
                '5': '0101',
                '6': '0110',
                '7': '0111',
                '8': '1000',
                '9': '1001',
                'a': '1010',
                'b': '1011',
                'c': '1100',
                'd': '1101',
                'e': '1110',
                'f': '1111',
            }
        self.assertEqual(diction, self.tab.hex_to_bin_dict())



import attr
from attr.validators import instance_of
import pandas as pd
import sys


@attr.s(slots=True)
class PhotonDF(object):
    """
    Create initial photon dataframe and set the channel as its index
    """

    dict_of_data = attr.ib(validator=instance_of(dict))
    num_of_channels = attr.ib(default=1, validator=instance_of(int))

    def gen_df(self):
        """
        If a single PMT channel exists, create a df_photons object.
        Else, concatenate the two data channels into a single dataframe.

        :return pd.DataFrame: Photon data
        """

        assert "Channel" in self.dict_of_data["PMT1"].columns

        try:
            df_photons = pd.concat(
                [self.dict_of_data["PMT1"].copy(), self.dict_of_data["PMT2"].copy()],
                axis=0,
            )
            self.num_of_channels = 2
        except KeyError:
            df_photons = self.dict_of_data["PMT1"].copy()
        except:
            print("Unknown error: ", sys.exc_info()[0])
        finally:
            df_photons.loc[:, "Channel"] = df_photons.loc[:, "Channel"].astype(
                "category"
            )
            df_photons.set_index(keys="Channel", inplace=True)

        return df_photons



"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.nd_hist_generator.line_signal_validators.validation_tools import SignalValidator
from pysight.nd_hist_generator.line_signal_validators.scanimage import ScanImageLineValidator
import pandas as pd
import numpy as np


class TestScanImage(TestCase):
    """
    Tests for the ScanImageValidator class
    """
    dict_of_data = dict(PMT1=pd.DataFrame([1, 10, 20, 30], columns=['abs_time']),
                        Lines=pd.DataFrame([0, 5, 10, 15, 20, 25, 30, 35]), columns=['abs_time'])
    vlad = SignalValidator(dict_of_data)

    def test_pairwise(self):
        dict_of_data = {}
        iter = ScanImageLineValidator(SignalValidator(dict_of_data))._ScanImageLineValidator__pairwise([1, 2, 3, 4])
        self.assertSequenceEqual([(1, 2), (2, 3), (3, 4)], list(iter))




"""
__author__ = Hagai Hargil
"""

import attr
from attr.validators import instance_of
from pysight.nd_hist_generator.movie import trunc_end_of_file
import numpy as np
import warnings
import h5py_cache
import os


@attr.s(slots=True)
class OutputParser(object):
    """
    Parse the wanted outputs and produce a dictionary with
    file pointers to the needed outputs
    """

    output_dict = attr.ib(validator=instance_of(dict))
    filename = attr.ib(validator=instance_of(str), converter=trunc_end_of_file)
    num_of_frames = attr.ib(default=1, validator=instance_of(int))
    x_pixels = attr.ib(default=512, validator=instance_of(int))
    y_pixels = attr.ib(default=512, validator=instance_of(int))
    z_pixels = attr.ib(default=1, validator=instance_of(int))
    num_of_channels = attr.ib(default=1, validator=instance_of(int))
    flim = attr.ib(default=False, validator=instance_of(bool))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    reprate = attr.ib(default=80e6, validator=instance_of(float))
    lst_metadata = attr.ib(default={}, validator=instance_of(dict))
    file_pointer_created = attr.ib(default=True, validator=instance_of(bool))
    cache_size = attr.ib(default=10 * 1024 ** 3, validator=instance_of(int))
    debug = attr.ib(default=False, validator=instance_of(bool))
    #: Dictionary with data - either the full or summed stack, with a list of HDF5 objects as channels
    outputs = attr.ib(init=False)
    #: Tuple of (num_of_frames, x, y, z, tau)
    data_shape = attr.ib(init=False)

    def run(self):
        """ Parse what the user required, creating a list of HDF5 dataset pointers for each channel """
        self.outputs = {}
        self.data_shape = self.determine_data_shape_full()
        if not self.output_dict:
            return
        try:
            self.outputs["memory"] = self.output_dict["memory"]
        except KeyError:
            pass
        f = self.__create_prelim_file()
        if f is not None:
            self.__populate_hdf(f)

    def __create_prelim_file(self):
        """ Try to create a preliminary .hdf5 file. Cache improves IO performance """
        if "stack" in self.output_dict or "summed" in self.output_dict:
            try:
                split = os.path.splitext(self.filename)[0]
                debugged = "_DEBUG" if self.debug else ""
                fullfile = f"{split + debugged}.hdf5"
                f = h5py_cache.File(
                    fullfile,
                    "w",
                    chunk_cache_mem_size=self.cache_size,
                    libver="latest",
                    w0=1,
                )
                self.outputs["filename"] = fullfile
            except PermissionError or OSError:
                self.file_pointer_created = False
                warnings.warn("Permission Error: Couldn't write data to disk.")
                return
            return f
        else:
            return

    def __populate_hdf(self, f):
        """
        Generate files and add metadata to each group, write out the data in chunks
        f: HDF5 file pointer
        """
        data_shape_summed = self.data_shape[1:]
        chunk_shape = list(self.data_shape)
        chunk_shape[0] = 1
        if "stack" in self.output_dict:
            try:
                self.outputs["stack"] = [
                    f.require_group("Full Stack").require_dataset(
                        name=f"Channel {channel}",
                        shape=self.data_shape,
                        dtype=np.uint8,
                        chunks=tuple(chunk_shape),
                        compression="gzip",
                    )
                    for channel in range(1, self.num_of_channels + 1)
                ]

                for key, val in self.lst_metadata.items():
                    for chan in range(self.num_of_channels):
                        self.outputs["stack"][chan].attrs.create(
                            name=key, data=val.encode()
                        )

            except PermissionError or OSError:
                self.file_pointer_created = False
        if "summed" in self.output_dict:
            try:
                self.outputs["summed"] = [
                    f.require_group("Summed Stack").require_dataset(
                        name=f"Channel {channel}",
                        shape=data_shape_summed,
                        dtype=np.uint16,
                        chunks=True,
                        compression="gzip",
                    )
                    for channel in range(1, self.num_of_channels + 1)
                ]
                for key, val in self.lst_metadata.items():
                    for chan in range(self.num_of_channels):
                        self.outputs["summed"][chan].attrs.create(
                            name=key, data=val.encode()
                        )

            except PermissionError or OSError:
                self.file_pointer_created = False

        f.close()
        if self.file_pointer_created is False:
            warnings.warn("Permission Error: Couldn't write data to disk.")

    @property
    def bins_bet_pulses(self) -> int:
        if self.flim:
            return int(np.ceil(1 / (self.reprate * self.binwidth)))
        else:
            return 1

    def determine_data_shape_full(self):
        """
        Return the tuple that describes the shape of the final dataset.
        Dimension order: [FRAME, X, Y, Z, LIFETIME]
        """
        non_squeezed = (
            self.x_pixels,
            self.y_pixels,
            self.z_pixels,
            self.bins_bet_pulses,
        )
        shape = tuple([dim for dim in non_squeezed if dim != 1])
        return (self.num_of_frames,) + shape  # we never "squeeze" the number of frames



import attr
from attr.validators import instance_of
import numpy as np
import pandas as pd
from typing import List, Dict, Tuple


@attr.s(slots=True)
class FrameChunk:
    """
    Holds a chunk of data and can histogram ir efficiently.
    Composed out of a Movie - any attribute\method not defined here is
    taken from the self.movie object with the __getattr__ method
    """

    movie = attr.ib()
    df_dict = attr.ib(validator=instance_of(dict), repr=False)
    frames = attr.ib(validator=instance_of(pd.Series))
    lines = attr.ib(validator=instance_of(pd.Series))
    frames_per_chunk = attr.ib(validator=instance_of(int))
    hist_dict = attr.ib(init=False)
    end_time = attr.ib(init=False)

    def __getattr__(self, item):
        return getattr(self.movie, item)

    def __attrs_post_init__(self):
        if len(self.frames) > 1:
            dif = np.uint64(self.frames.diff().mean())
            self.end_time = self.frames.iloc[-1] + dif
        else:
            self.end_time = self.frames.iloc[0] + 1

    def create_hist(self) -> Dict[int, Tuple[np.ndarray]]:
        """
        Main method to create the histogram of data. Assigns each event
        in the dataframe to its correct location, for each channel.

        :return: A dictionary with its keys being the spectral channels
                of the data, and the value is a tuple of the histogrammed data and the edges.
        """
        self.hist_dict: Dict[int, Tuple[np.ndarray]] = {}
        for chan in self.df_dict:
            list_of_edges = self.__create_hist_edges(chan)
            data_columns = []
            data_columns.append(self.df_dict[chan]["abs_time"].values)
            data_columns.append(self.df_dict[chan]["time_rel_line"].values)
            try:
                data_columns.append(self.df_dict[chan]["Phase"].values)
            except KeyError:
                pass
            try:
                data_columns.append(self.df_dict[chan]["time_rel_pulse"].values)
            except KeyError:
                pass

            hist, edges = np.histogramdd(sample=data_columns, bins=list_of_edges)
            hist = hist.astype(np.uint8).reshape(
                (self.frames_per_chunk,) + self.data_shape[1:]
            )

            if self.bidir:
                hist[:, 1::2, ...] = np.flip(hist[:, 1::2, ...], axis=2)
            if self.censor:
                hist = self.__censor_correction(hist)

            self.hist_dict[chan] = hist, edges
        return self.hist_dict

    def __create_hist_edges(self, chan) -> List[np.ndarray]:
        """
        Generate the grid of the histogram.

        :param int chan: Channel number

        :return ``list`` of ``np.ndarray``: One for each dimension
        """
        edges = []
        edges.append(self.__create_line_edges())
        edges.append(self.__create_col_edges())

        if "Phase" in self.df_dict[chan].columns:
            edges.append(self.__linspace_along_sine())

        if "time_rel_pulse" in self.df_dict[chan].columns:
            edges.append(self.__create_laser_edges())

        return edges

    def __censor_correction(self, hist):
        raise NotImplementedError(
            "No censor correction as of yet. Contact package authors."
        )

    def __create_frame_edges(self) -> np.ndarray:
        """ Create edges for a numpy histogram for the frames dimension """

        assert self.frames.shape[0] == self.frames_per_chunk
        frames = np.hstack((self.frames.values, self.frames.values[-1] + np.uint64(1)))
        return frames

    def __create_line_edges(self) -> np.ndarray:
        """ Takes existing lines and turns them into bin edges. """

        assert (
            self.lines.shape[0] <= self.x_pixels * self.frames_per_chunk
        )  # last chunk can have less frames
        all_lines = np.hstack(
            (self.lines.values, self.lines.values[-1] + self.line_delta)
        )
        return all_lines

    def __create_col_edges(self) -> np.ndarray:
        if self.x_pixels == 1:
            return np.linspace(
                0, self.end_time, num=self.y_pixels + 1, endpoint=True, dtype=np.uint64
            )

        delta = self.line_delta if self.bidir else self.line_delta / 2
        start = 0
        if self.image_soft == "ScanImage":
            start = delta * 0.04
        col_end = delta * self.fill_frac / 100 if self.fill_frac > 0 else delta

        return np.linspace(
            start=start,
            stop=int(col_end),
            num=self.y_pixels + 1,
            endpoint=True,
            dtype=np.uint64,
        )

    def __linspace_along_sine(self) -> np.ndarray:
        """
        Find the points that are evenly spaced along a sine function between pi/2 and 3*pi/2

        :return: Array of bin edges
        """
        lower_bound = -1 if self.tag_as_phase else 0
        upper_bound = 1 if self.tag_as_phase else self.tag_period
        pts = []
        relevant_idx = []

        bin_edges = np.linspace(
            lower_bound, upper_bound, self.z_pixels + 1, endpoint=True
        )[:, np.newaxis]
        dx = 0.00001
        x = np.arange(np.pi / 2, 3 * np.pi / 2 + dx, step=dx, dtype=np.float64)
        sinx = np.sin(x)
        locs = np.where(np.isclose(sinx, bin_edges, atol=1e-05))
        vals, first_idx, count = np.unique(
            locs[0], return_index=True, return_counts=True
        )
        assert len(vals) == len(bin_edges)
        for first_idx, count in zip(first_idx, count):
            idx_to_append = locs[1][first_idx + count // 2]
            relevant_idx.append(idx_to_append)
            pts.append(sinx[idx_to_append])

        return np.array(pts)



"""
#!/usr/bin/env python
# -*- coding: utf-8 -*-
__author__ = Hagai Har-Gil
Created on Thu Oct 13 09:37:02 2016
"""
from typing import Union, Tuple
import matplotlib

matplotlib.rcParams["backend"] = "TkAgg"
import json
import pandas as pd
import warnings
import colorama
import numpy as np
import pickle
import matplotlib.pyplot as plt

from pysight.ascii_list_file_parser.file_io import FileIO
from pysight.ascii_list_file_parser.tabulation import Tabulate
from pysight.nd_hist_generator.allocation import Allocate
from pysight.nd_hist_generator.movie import Movie
from pysight.ascii_list_file_parser import timepatch_switch
from pysight.nd_hist_generator.outputs import OutputParser
from pysight.nd_hist_generator.gating import GatedDetection
from pysight.nd_hist_generator.photon_df import PhotonDF
from pysight.nd_hist_generator.tag_bits import ParseTAGBits
from pysight.ascii_list_file_parser.distribute_data import DistributeData
from pysight.nd_hist_generator.line_signal_validators.validation_tools import (
    SignalValidator,
)
from pysight.tkinter_gui_multiscaler import GuiAppLst
from pysight.tkinter_gui_multiscaler import verify_gui_input
from pysight.nd_hist_generator.volume_gen import VolumeGenerator
from pysight.binary_list_file_parser.binary_parser import BinaryDataParser

colorama.init()


def main_data_readout(gui):
    """
    Main function that reads the lst file and processes its data.
    Should not be run independently - only from other "run_X" functions.
    """
    # Read the file
    if gui.filename.endswith(".lst"):
        cur_file = FileIO(
            filename=gui.filename,
            debug=gui.debug,
            input_start=gui.input_start,
            input_stop1=gui.input_stop1,
            input_stop2=gui.input_stop2,
            binwidth=gui.binwidth,
            use_sweeps=gui.sweeps_as_lines,
            mirror_phase=gui.phase,
        )
        cur_file.run()

        if cur_file.is_binary:
            binary_parser = BinaryDataParser(
                data=cur_file.data,
                data_range=cur_file.data_range,
                timepatch=cur_file.timepatch,
                use_tag_bits=gui.tag_bits,
                dict_of_inputs=cur_file.dict_of_input_channels,
            )
            binary_parser.run()
        else:
            # Create input structures and create a DataFrame
            dict_of_slices_hex = timepatch_switch.ChoiceManagerHex().process(
                cur_file.timepatch
            )
            tabulated_data = Tabulate(
                data_range=cur_file.data_range,
                data=cur_file.data,
                dict_of_inputs=cur_file.dict_of_input_channels,
                use_tag_bits=gui.tag_bits,
                dict_of_slices_hex=dict_of_slices_hex,
                time_after_sweep=cur_file.time_after,
                acq_delay=cur_file.acq_delay,
                num_of_channels=cur_file.num_of_channels,
            )
            tabulated_data.run()

            separated_data = DistributeData(
                df=tabulated_data.df_after_timepatch,
                dict_of_inputs=tabulated_data.dict_of_inputs,
                use_tag_bits=gui.tag_bits,
            )
            separated_data.run()

    ####### START OF "PUBLIC API" ##########
    try:
        if cur_file.is_binary:
            relevant_columns = binary_parser.data_to_grab
            dict_of_data = binary_parser.dict_of_data
        else:
            relevant_columns = separated_data.data_to_grab
            dict_of_data = separated_data.dict_of_data
        lst_metadata = cur_file.lst_metadata
    except NameError:
        with open(gui.filename, "rb") as f:
            dict_of_data = pickle.load(f)
        lst_metadata = dict()
        relevant_columns = ["abs_time"]

    validated_data = SignalValidator(
        dict_of_data=dict_of_data,
        num_of_frames=gui.num_of_frames,
        binwidth=float(gui.binwidth),
        use_sweeps=gui.sweeps_as_lines,
        delay_between_frames=float(gui.frame_delay),
        data_to_grab=relevant_columns,
        line_freq=gui.line_freq,
        num_of_lines=gui.x_pixels,
        bidir=gui.bidir,
        bidir_phase=gui.phase,
        image_soft=gui.imaging_software,
    )

    validated_data.run()

    photon_df = PhotonDF(dict_of_data=validated_data.dict_of_data)
    tag_bit_parser = ParseTAGBits(
        dict_of_data=validated_data.dict_of_data,
        photons=photon_df.gen_df(),
        use_tag_bits=gui.tag_bits,
        bits_dict=gui.tag_bits_dict,
    )

    analyzed_struct = Allocate(
        bidir=gui.bidir,
        tag_offset=gui.tag_offset,
        laser_freq=float(gui.reprate),
        binwidth=float(gui.binwidth),
        tag_pulses=int(gui.tag_pulses),
        phase=gui.phase,
        keep_unidir=gui.keep_unidir,
        flim=gui.flim,
        censor=gui.censor,
        dict_of_data=validated_data.dict_of_data,
        df_photons=tag_bit_parser.gen_df(),
        tag_freq=float(gui.tag_freq),
        tag_to_phase=True,
    )
    analyzed_struct.run()

    # Determine type and shape of wanted outputs, and open the file pointers there
    outputs = OutputParser(
        num_of_frames=len(validated_data.dict_of_data["Frames"]),
        output_dict=gui.outputs,
        filename=gui.filename,
        x_pixels=gui.x_pixels,
        y_pixels=gui.y_pixels,
        z_pixels=gui.z_pixels if analyzed_struct.tag_interp_ok else 1,
        num_of_channels=analyzed_struct.num_of_channels,
        flim=gui.flim,
        binwidth=gui.binwidth,
        reprate=gui.reprate,
        lst_metadata=lst_metadata,
        debug=gui.debug,
    )
    outputs.run()

    if gui.gating:
        gated = GatedDetection(
            raw=analyzed_struct.df_photons, reprate=gui.reprate, binwidth=gui.binwidth
        )
        gated.run()

    data_for_movie = gated.data if gui.gating else analyzed_struct.df_photons

    # Create a movie object
    volume_chunks = VolumeGenerator(
        frames=analyzed_struct.dict_of_data["Frames"], data_shape=outputs.data_shape
    )
    frame_slices = volume_chunks.create_frame_slices()

    final_movie = Movie(
        data=data_for_movie,
        frames=analyzed_struct.dict_of_data["Frames"],
        frame_slices=frame_slices,
        num_of_frame_chunks=volume_chunks.num_of_chunks,
        reprate=float(gui.reprate),
        name=gui.filename,
        data_shape=outputs.data_shape,
        binwidth=float(gui.binwidth),
        bidir=gui.bidir,
        fill_frac=gui.fill_frac
        if cur_file.fill_fraction == -1.0
        else cur_file.fill_fraction,
        outputs=outputs.outputs,
        censor=gui.censor,
        mirror_phase=gui.phase,
        lines=analyzed_struct.dict_of_data["Lines"],
        num_of_channels=analyzed_struct.num_of_channels,
        flim=gui.flim,
        lst_metadata=cur_file.lst_metadata,
        exp_params=analyzed_struct.exp_params,
        line_delta=int(validated_data.line_delta),
        use_sweeps=gui.sweeps_as_lines,
        tag_as_phase=True,
        tag_freq=float(gui.tag_freq),
        image_soft=gui.imaging_software,
        frames_per_chunk=volume_chunks.frames_per_chunk,
    )

    final_movie.run()

    return analyzed_struct.df_photons, final_movie


class GUIClass:
    """ Helper class to create intermediate representation of the GUI's content """

    def __init__(self, **entries):
        self.__dict__.update(entries)
        self.tag_bits_dict = dict(
            bits_grp_1_label=entries["bits_grp_1_label"],
            bits_grp_2_label=entries["bits_grp_2_label"],
            bits_grp_3_label=entries["bits_grp_3_label"],
        )

    @property
    def outputs(self):
        """ Create a dictionary with the wanted user outputs. """
        output = {}

        if self.summed is True:
            output["summed"] = True
        if self.memory is True:
            output["memory"] = True
        if self.stack is True:
            output["stack"] = True

        if "stack" in output:
            if not "summed" in output and not "memory" in output:
                warnings.warn(
                    "Performance Warning: Writing data to file might take a long time when the required"
                    " output is only 'Full Stack'."
                )
        return output


def tkinter_to_object(gui: Union[GuiAppLst, dict]) -> GUIClass:
    """ Convert a tkinter instance into a pickable dictionary """
    if isinstance(gui, dict):
        gui["tuple_of_data_sources"] = (
            "PMT1",
            "PMT2",
            "Lines",
            "Frames",
            "Laser",
            "TAG Lens",
            "Empty",
        )
        return GUIClass(**gui)
    else:
        dic = {
            key: val.get() for key, val in gui.__dict__.items() if "Var" in repr(val)
        }
        dic["tuple_of_data_sources"] = gui.tuple_of_data_sources
        return GUIClass(**dic)


def convert_json_to_input_dict(cfg_fname):
    """ Convert a config file to a usable dictionary """
    if not isinstance(cfg_fname, str):
        raise TypeError

    try:
        with open(cfg_fname) as f:
            gui = json.load(f)
        gui = {key: val[1] for key, val in gui.items()}
        return gui
    except:
        raise TypeError


def run(cfg_file: str = None) -> Tuple[pd.DataFrame, Movie]:
    """ Run PySight.

    :param str cfg_file: Optionally supply an existing configuration filename. Otherwise a GUI will open.

    :return (pd.DataFrame, Movie): DataFrame with all the data, and a ``Movie`` object.
    """
    from pysight.tkinter_gui_multiscaler import GuiAppLst
    from pysight.tkinter_gui_multiscaler import verify_gui_input

    if cfg_file:
        gui = convert_json_to_input_dict(cfg_file)
    else:
        gui = GuiAppLst()
        gui.root.mainloop()
    gui_as_object = tkinter_to_object(gui)
    verify_gui_input(gui_as_object)
    return main_data_readout(gui_as_object)


def run_batch_lst(
    foldername: str,
    glob_str: str = "*.lst",
    recursive: bool = False,
    cfg_file: str = "",
) -> pd.DataFrame:
    """
    Run PySight on all list files in the folder

    :param str foldername: - Main folder to run the analysis on.
    :param str glob_str: String for the `glob` function to filter list files
    :param bool recursive: Whether the search should be recursive.
    :param str cfg_file: Name of config file to use
    :return pd.DataFrame: Record of analyzed data
    """

    import pathlib
    from pysight.tkinter_gui_multiscaler import GuiAppLst
    from pysight.tkinter_gui_multiscaler import verify_gui_input
    import numpy as np

    path = pathlib.Path(foldername)
    num_of_files = 0
    if not path.exists():
        raise UserWarning(f"Folder {foldername} doesn't exist.")
    if recursive:
        all_lst_files = path.rglob(glob_str)
        print(f"Running PySight on the following files:")
        for file in list(all_lst_files):
            print(str(file))
            num_of_files += 1
        all_lst_files = path.rglob(glob_str)
    else:
        all_lst_files = path.glob(glob_str)
        print(f"Running PySight on the following files:")
        for file in list(all_lst_files):
            print(str(file))
            num_of_files += 1
        all_lst_files = path.glob(glob_str)

    data_columns = ["fname", "done", "error"]
    data_record = pd.DataFrame(
        np.zeros((num_of_files, 3)), columns=data_columns
    )  # store result of PySight
    try:
        cfg_dict = convert_json_to_input_dict(cfg_file)
        named_gui = tkinter_to_object(cfg_dict)
    except TypeError:
        gui = GuiAppLst()
        gui.root.mainloop()
        gui.filename.set(".lst")  # no need to choose a list file
        named_gui = tkinter_to_object(gui)
    verify_gui_input(named_gui)

    try:
        for idx, lst_file in enumerate(all_lst_files):
            named_gui.filename = str(lst_file)
            data_record.loc[idx, "fname"] = str(lst_file)
            try:
                main_data_readout(named_gui)
            except BaseException as e:
                print(f"File {str(lst_file)} returned an error. Moving onwards.")
                data_record.loc[idx, "done"] = False
                data_record.loc[idx, "error"] = repr(e)
            else:
                data_record.loc[idx, "done"] = True
                data_record.loc[idx, "error"] = None
    except TypeError as e:
        print(repr(e))

    print(f"Summary of batch processing:\n{data_record}")
    return data_record


def mp_batch(
    foldername, glob_str="*.lst", recursive=False, n_proc=None, cfg_file: str = ""
):
    """
    Run several instances of PySight using the multiprocessing module.

    :param str foldername: Folder to scan
    :param str glob_str: Glob string to filter files
    :param bool recursive: Whether to scan subdirectories as well
    :param int n_proc: Number of processes to use (None means all)
    :param str cfg_file: Configuration file name
    """
    import pathlib
    import multiprocessing as mp

    path = pathlib.Path(foldername)
    if not path.exists():
        raise UserWarning(f"Folder {foldername} doesn't exist.")
    if recursive:
        all_lst_files = path.rglob(glob_str)
    else:
        all_lst_files = path.glob(glob_str)
    print(f"Running PySight on the following files:")
    for file in list(all_lst_files):
        print(str(file))

    all_lst_files = path.rglob(glob_str) if recursive else path.glob(glob_str)
    try:
        gui = convert_json_to_input_dict(cfg_file)
    except TypeError:
        gui = GuiAppLst()
        gui.root.mainloop()
        gui.filename.set(".lst")  # no need to choose a list file
    g = tkinter_to_object(gui)
    verify_gui_input(g)
    all_guis = []
    for file in all_lst_files:
        g.filename = str(file)
        all_guis.append(g)
    pool = mp.Pool(n_proc)
    pool.map(main_data_readout, all_guis)


if __name__ == "__main__":
    df, movie = run()



"""
__author__ = Hagai Har-Gil
"""
from unittest import TestCase
import numpy as np
import pathlib
import h5py
import tkinter

from pysight.main import run


class TestEntirePipeline(TestCase):
    """ Integration tests for the entire application """

    def test_standard_pipeline(self):
        cfg_file = str(next(pathlib.Path('.').rglob('*jul.json')).absolute())
        data_file = str(next(pathlib.Path('.').rglob('*jul.hdf5')).absolute())
        df, movie = run(cfg_file)
        with h5py.File(data_file, 'r') as f:
            np.testing.assert_array_equal(movie.summed_mem[1], np.array(f['/Summed Stack/Channel 1']))

    def test_mscan_pipeline(self):
        cfg_file = str(next(pathlib.Path('.').rglob('*fly.json')).absolute())
        data_file = str(next(pathlib.Path('.').rglob('*fly.hdf5')).absolute())
        df, movie = run(cfg_file)
        with h5py.File(data_file, 'r') as f:
            np.testing.assert_array_equal(movie.summed_mem[1], np.array(f['/Summed Stack/Channel 1']))

    def test_tag_pipeline(self):
        cfg_file = str(next(pathlib.Path('.').rglob('*tag.json')).absolute())
        data_file = str(next(pathlib.Path('.').rglob('*tag.hdf5')).absolute())
        df, movie = run(cfg_file)
        with h5py.File(data_file, 'r') as f:
            np.testing.assert_array_equal(movie.summed_mem[1], np.array(f['/Summed Stack/Channel 1']))



from unittest import TestCase
from collections import namedtuple
from pprint import pprint

from pysight.nd_hist_generator.frame_chunk import *
from pysight.nd_hist_generator.movie import *


def gen_data_df(frame_num=10, line_num=100, end=1000, channels=2):
    """
    Mock data for tests.
    Returns:
        df - The full DataFrame
        frames only
        lines only
        x pixels
        y pixels
    """
    photons = np.arange(0, end, dtype=np.uint64)
    channel = np.ones_like(photons)
    if channels == 2:
        channel[len(channel)//2:] = 2
    lines = np.linspace(0, end, num=line_num, endpoint=False, dtype=np.uint64)
    x_pix = int(len(photons) / len(lines))
    ones_lines = np.ones((1, int(len(photons) / len(lines))),
                         dtype=np.uint64)
    frames = np.linspace(0, end, num=frame_num, dtype=np.uint64, endpoint=False)
    frames_ser = pd.Series(frames, index=frames)
    ones_frames = np.ones((1, int(len(photons) / len(frames))),
                          dtype=np.uint64)
    lines = (np.atleast_2d(lines).T @ ones_lines).ravel()
    frames = (np.atleast_2d(frames).T @ ones_frames).ravel()
    assert len(lines) == len(frames) == len(photons)

    df = pd.DataFrame({'abs_time': photons,
                       'time_rel_line': photons - lines,
                       'Lines': lines, 'Frames': frames,
                       'Channel': channel})
    df.set_index(['Channel', 'Frames', 'Lines'], drop=True, inplace=True)
    y_pix = x_pix
    lines_to_return = pd.Series(np.unique(lines), index=np.repeat(frames_ser, line_num//frame_num))

    return df, frames_ser, lines_to_return, x_pix, y_pix


class TestFrameChunk(TestCase):
    df, frames, lines, x, y = gen_data_df()
    movie_single = Movie(df, lines, outputs={'memory': True}, line_delta=int(lines.diff().mean()),
                         fill_frac=100., bidir=True, data_shape=(len(frames), x, y),
                         frame_slices=(slice(frame) for frame in frames), frames=frames,
                         frames_per_chunk=10)
    df_dict = {1: df.xs(key=(1, 100), level=('Channel', 'Frames'),
                        drop_level=False),
               2: df.xs(key=(2, 100), level=('Channel', 'Frames'),
                        drop_level=False)}
    chunk_single = FrameChunk(movie=movie_single, df_dict=df_dict, frames_per_chunk=10, frames=frames,
                              lines=lines)

    movie_multi = Movie(df, lines=lines, outputs={'memory': True},
                        line_delta=int(lines.diff().mean()), fill_frac=100., bidir=True,
                        data_shape=(len(frames), x, y), frames=frames,
                        frame_slices=(slice(frame) for frame in frames), frames_per_chunk=4)
    sl = pd.IndexSlice[slice(1), slice(100, 400)]
    chunk_multi = FrameChunk(movie=movie_multi, df_dict={1: df.loc[sl, :]}, frames_per_chunk=4,
                             frames=frames, lines=lines)

    def test_frame_edges_single_chunk(self):
        fr = self.chunk_single._FrameChunk__create_frame_edges()
        np.testing.assert_equal(fr, np.array([0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 901], dtype=np.uint64))

    def test_frame_edges_multiple_frames(self):
        sl = pd.IndexSlice[slice(1), slice(100, 400)]
        movie = Movie(self.df, self.lines, outputs={'memory': True},
                      line_delta=int(self.lines.diff().mean()), fill_frac=100., bidir=True,
                      data_shape=(len(self.frames), self.x, self.y), frames=self.frames,
                      frame_slices=(slice(frame) for frame in self.frames), frames_per_chunk=4, )
        chunk_multi = FrameChunk(movie=movie, df_dict={1: self.df.loc[sl, :]},
                                 frames=pd.Series([100, 200, 300, 400], dtype=np.uint64),
                                 lines=pd.Series([10]), frames_per_chunk=4)
        fr = chunk_multi._FrameChunk__create_frame_edges()
        np.testing.assert_equal(fr, np.array([100, 200, 300, 400, 401]))

    def test_line_edges_single_chunk(self):
        li = self.chunk_single._FrameChunk__create_line_edges()
        lines = np.arange(0, 1010, 10)
        np.testing.assert_equal(li, lines)

    def test_line_edges_multi_chunk(self):
        sl = pd.IndexSlice[slice(1), slice(100, 400)]
        movie = Movie(self.df, self.lines, outputs={'memory': True},
                      line_delta=int(self.lines.diff().mean()), fill_frac=100., bidir=True,
                      data_shape=(len(self.frames), self.x, self.y), frames=self.frames,
                      frame_slices=(slice(frame) for frame in self.frames), frames_per_chunk=4, )
        chunk_multi = FrameChunk(movie=movie, df_dict={1: self.df.loc[sl, :]}, frames=self.frames,
                                 lines=self.lines.loc[slice(100, 400)], frames_per_chunk=4, )
        li = chunk_multi._FrameChunk__create_line_edges()
        lines = np.arange(100, 510, 10)
        np.testing.assert_equal(li, lines)

    def test_col_edges_single_frame(self):
        cr = self.chunk_single._FrameChunk__create_col_edges()
        cols = np.arange(11)
        np.testing.assert_equal(cr, cols)

    def test_col_edges_multi_frame(self):
        cr = self.chunk_multi._FrameChunk__create_col_edges()
        cols = np.arange(11)
        self.assertSequenceEqual(cr.tolist(), cols.tolist())

    def test_linspace_along_sine_1_pix_z(self):
        sine = self.chunk_single._FrameChunk__linspace_along_sine()
        np.testing.assert_almost_equal(np.array([-0.99999503, 0.99999501]),
                                       sine.tolist())

    def test_linspace_along_sine_100_pix_z(self):
        movie_for_sine = Movie(self.df, self.lines, data_shape=(1, 512, 512, 100),
                               frame_slices=(1,), frames=self.frames)
        chunk = FrameChunk(movie_for_sine, self.df_dict, frames_per_chunk=1,
                           frames=self.frames, lines=self.lines)
        sin = chunk._FrameChunk__linspace_along_sine()
        true = np.array([-9.99995030e-01, -9.80000436e-01, -9.60000408e-01, -9.40001149e-01,
                         -9.20001249e-01, -9.00001813e-01, -8.80000700e-01, -8.60002042e-01,
                         -8.40000245e-01, -8.20001519e-01, -8.00005073e-01, -7.80004916e-01,
                         -7.60000363e-01, -7.40002230e-01, -7.20000940e-01, -7.00004411e-01,
                         -6.80000761e-01, -6.60003690e-01, -6.40004155e-01, -6.20000761e-01,
                         -6.00002052e-01, -5.80004059e-01, -5.60006523e-01, -5.40005525e-01,
                         -5.20002325e-01, -5.00004241e-01, -4.79999088e-01, -4.59998646e-01,
                         -4.40004490e-01, -4.19998505e-01, -3.99997092e-01, -3.79997574e-01,
                         -3.59996063e-01, -3.39996968e-01, -3.20003966e-01, -3.00000972e-01,
                         -2.79999581e-01, -2.60001420e-01, -2.39997886e-01, -2.19999222e-01,
                         -2.00005636e-01, -1.80007104e-01, -1.60002981e-01, -1.40002236e-01,
                         -1.20003763e-01, -1.00006221e-01, -8.00080672e-02, -6.00076010e-02,
                         -4.00029965e-02, -2.00023392e-02, -3.67321539e-06,  1.99949942e-02,
                          3.99956559e-02,  6.00002679e-02,  8.00007444e-02,  9.99989111e-02,
                          1.19996470e-01,  1.39994962e-01,  1.59995730e-01,  1.79999878e-01,
                          1.99998438e-01,  2.20001811e-01,  2.40000462e-01,  2.59994326e-01,
                          2.80002129e-01,  2.99993964e-01,  3.19997006e-01,  3.39999463e-01,
                          3.59998538e-01,  3.80000028e-01,  3.99999524e-01,  4.20000914e-01,
                          4.39997893e-01,  4.60001002e-01,  4.80001416e-01,  4.99997879e-01,
                          5.19996050e-01,  5.39999341e-01,  5.60000436e-01,  5.79998074e-01,
                          5.99996174e-01,  6.19994997e-01,  6.39998510e-01,  6.59998171e-01,
                          6.79995375e-01,  6.99999165e-01,  7.20002781e-01,  7.39997288e-01,
                          7.59995588e-01,  7.80000319e-01,  7.99994665e-01,  8.19997314e-01,
                          8.40001684e-01,  8.59998293e-01,  8.79997210e-01,  8.99998610e-01,
                          9.19998370e-01,  9.39998643e-01,  9.59998351e-01,  9.79998974e-01,
                          9.99995007e-01])
        np.testing.assert_almost_equal(sin, true)



from typing import Dict, List
from attr.validators import instance_of
import numpy as np
import attr
import re
from enum import Enum


class LstFormat(Enum):
    ASCII = "ascii"
    BINARY = "binary"


@attr.s(slots=True)
class FileIO(object):
    """
    Manage pipeline of file IO process.

    :param str filename: List file name
    :param bool debug: Run a debug build (limited number of lines for quick execution of the entire pipeline)
    :param str input_start: Data type in analog channel 'START' (6)
    :param str input_stop1: Data type in analog channel 'STOP1' (1)
    :param str input_stop2: Data type in analog channel 'STOP2' (2)
    :param float binwidth: Multiscaler resolution in seconds (100-800 picoseconds)
    :param bool use_sweeps: Use the sweeps counter as a new frame indicator (dev mode)
    """

    filename = attr.ib(validator=instance_of(str))
    debug = attr.ib(default=False, validator=instance_of(bool))
    input_start = attr.ib(default="Frames", validator=instance_of(str))
    input_stop1 = attr.ib(default="PMT1", validator=instance_of(str))
    input_stop2 = attr.ib(default="Lines", validator=instance_of(str))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    use_sweeps = attr.ib(default=False, validator=instance_of(bool))
    mirror_phase = attr.ib(default=-2.78, validator=instance_of(float))
    is_binary = attr.ib(init=False)
    timepatch = attr.ib(init=False)
    data_range = attr.ib(init=False)
    time_after = attr.ib(init=False)
    acq_delay = attr.ib(init=False)
    start_of_data_pos = attr.ib(init=False)
    dict_of_input_channels = attr.ib(init=False)
    list_of_recorded_data_channels = attr.ib(init=False)
    data = attr.ib(init=False)
    lst_metadata = attr.ib(init=False)
    fill_fraction = attr.ib(init=False)
    num_of_channels = attr.ib(init=False)

    def run(self):
        """ Open file and find the needed parameters """
        self.determine_binary()
        metadata: str = self.__get_metadata()
        self.timepatch: str = self.get_timepatch(metadata)
        if self.timepatch == "3" and not self.is_binary:
            raise NotImplementedError(
                'Timepatch value "3" is currently not supported for hex files. Please message the package owner.'
            )
        self.data_range: int = self.get_range(metadata)

        self.start_of_data_pos: int = self.get_start_pos()
        self.time_after: int = self.__get_hold_after(cur_str=metadata)
        self.acq_delay: int = self.__get_fstchan(cur_str=metadata)
        self.fill_fraction: float = self.__calc_actual_fill_fraction()
        self.dict_of_input_channels: dict = self.create_inputs_dict()
        self.list_of_recorded_data_channels: list = self.find_active_channels(metadata)
        self.compare_recorded_and_input_channels()
        num_of_items = self.determine_num_of_items()
        # Grab some additional metadata from the file, to be saved later
        self.__parse_extra_metadata(metadata)
        # Read the actual data
        self.data: np.ndarray = self.read_lst(num_of_items=num_of_items)
        print("File read. Sorting the file according to timepatch...")

    def __get_metadata(self) -> str:
        """
        Read the file's metadata to be parsed by other functions.

        :return str: String with the .lst file's metadata
        """
        file_mode = "rb" if self.is_binary else "r"
        with open(self.filename, file_mode) as f:
            metadata = f.read(5000)

        return metadata

    def determine_num_of_items(self):
        if self.debug == False:
            num_of_items = -1
            read_string = f'Reading file "{self.filename}"...'
        else:
            num_of_items = 0.2e6
            read_string = f'[DEBUG] Reading file "{self.filename}"...'

        print(read_string)
        return num_of_items

    @staticmethod
    def create_data_length_dict():
        """  CURRENTLY DEPRECATED """
        dict_of_data_length = {
            "0": 16,
            "5": 32,
            "1": 32,
            "1a": 48,
            "2a": 48,
            "22": 48,
            "32": 48,
            "2": 48,
            "5b": 64,
            "Db": 64,
            "f3": 64,
            "43": 64,
            "c3": 64,
            "3": 64,
        }

        return dict_of_data_length

    @staticmethod
    def hex_to_bin_dict() -> Dict:
        """ Create a simple dictionary that maps a hex input into a 4 letter binary output. """
        diction = {
            "0": "0000",
            "1": "0001",
            "2": "0010",
            "3": "0011",
            "4": "0100",
            "5": "0101",
            "6": "0110",
            "7": "0111",
            "8": "1000",
            "9": "1001",
            "a": "1010",
            "b": "1011",
            "c": "1100",
            "d": "1101",
            "e": "1110",
            "f": "1111",
        }
        return diction

    def determine_binary(self):
        """
        Determine whether we're dealing with a binary or a hex file.
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        try:
            with open(self.filename, "r") as f:
                txt = f.read(400)
        except FileNotFoundError:
            raise FileNotFoundError(f"File {self.filename} doesn't exist.")
        except UnicodeDecodeError:
            with open(self.filename, "rb") as f:
                txt = f.read(400).decode()
        except:
            raise Exception(f"File {self.filename} read unsuccessfully.")

        reg = re.compile(r"\nmpafmt=(\w{3})")
        match = reg.findall(txt)
        if len(match) == 0:
            raise RuntimeError("Could not resolve file type.")
        if match[0] == "dat":
            self.is_binary = True
        elif match[0] == "asc":
            self.is_binary = False

    def get_range(self, cur_str) -> int:
        """
        Finds the "range" of the current file in the proper units
        :return int: range as defined by MCS6A
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str = b"range=(\d+)"
        else:
            format_str = r"range=(\d+)"

        format_range = re.compile(format_str)
        range_lst = int(re.search(format_range, cur_str).group(1))

        return range_lst

    def get_timepatch(self, cur_str: str) -> str:
        """
        Get the time patch value out of of a list file.

        :param cur_str: Start of file to be analyzed.
        :return str: Time patch value as string.
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str: str = b"time_patch=(\w+)"
        else:
            format_str: str = r"time_patch=(\w+)"

        format_timepatch = re.compile(format_str)
        timepatch = re.search(format_timepatch, cur_str).group(1)
        try:
            timepatch = timepatch.decode("utf-8")
        except AttributeError:
            assert isinstance(timepatch, str)
        finally:
            if self.is_binary and timepatch in ("1a", "2a", "22", "32", "2"):
                raise NotImplementedError(
                    f"The timepatch used ({timepatch}) isn't supported "
                    "for binary files since it uses a 6-byte word representation. "
                    "Please disallow this option in the MPANT software."
                )
            return timepatch

    def find_active_channels(self, cur_str) -> List[bool]:
        """
        Create a dictionary containing the active channels.

        :param cur_str: String to be analyzed.
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str = b"active=(\d)"
            match = b"1"
        else:
            format_str = r"active=(\d)"
            match = "1"

        format_active = re.compile(format_str)
        active_channels: List[bool] = [False, False, False, False, False, False]
        list_of_matches = re.findall(format_active, cur_str)

        for idx, cur_match in enumerate(list_of_matches):
            if cur_match == match:
                active_channels[idx] = True

        return active_channels

    def get_start_pos(self) -> int:
        """
        Returns the start position of the data

        :return int: Integer of file position for f.seek() method
        """
        if self.filename == "":
            raise ValueError("No filename given.")

        if self.is_binary:
            format_str = b"DATA]\r\n"
            file_mode = "rb"
        else:
            format_str = r"DATA]\n"
            file_mode = "r"

        format_data = re.compile(format_str)
        pos_in_file: int = 0
        line_num: int = 0
        with open(self.filename, file_mode) as f:
            while pos_in_file == 0 and line_num < 1000:
                line = f.readline()
                line_num += 1
                match = re.search(format_data, line)
                if match is not None:
                    pos_in_file = f.tell()
                    return pos_in_file  # to have the [DATA] as header

    def __get_hold_after(self, cur_str) -> int:
        """
        Read the time (in ns, and convert to timebins) that is considered a
        "hold after", or "hold-off", after a single sweep. Add that time, along with a 96 ns
        inherit delay, to all future times of the sweep.

        :param cur_str: String to parse
        :return int: Final time that has to be added to all sweeps, in timebins
        """
        if self.is_binary:
            format_str: str = b"holdafter=([\w\+]+)"
        else:
            format_str: str = r"holdafter=([\w\+]+)"

        format_holdafter = re.compile(format_str)
        holdafter = float(re.search(format_holdafter, cur_str).group(1))
        EOS_DEADTIME = 96  # ns, current spec of Multiscaler

        time_after_sweep = int((EOS_DEADTIME + holdafter) * 10 ** (-9) / self.binwidth)
        return time_after_sweep

    def __get_fstchan(self, cur_str: str) -> int:
        """
        Read the acquisition delay of each sweep, called "fstchan" in the list files

        :param str cur_str: Metadata to be parsed
        :return int: Acq delay in timebins
        """
        if self.is_binary:
            format_str: str = b"fstchan=(\w+)"
        else:
            format_str: str = r"fstchan=(\w+)"

        NANOSECONDS_PER_FSTCHAN = 6.4e-9  # Multiscaler const parameter

        format_fstchan = re.compile(format_str)
        fstchan = (
            int(re.search(format_fstchan, cur_str).group(1)) * NANOSECONDS_PER_FSTCHAN
        )  # in nanoseconds
        acq_delay = int(fstchan / self.binwidth)
        return acq_delay

    def read_lst(self, num_of_items: int = -1) -> np.ndarray:
        """
        Updated version of LST readout using array slicing (and not Pandas slicing).

        :param int num_of_items: Number of lines to read. -1 is all file.
        """

        if self.filename is "" or self.start_of_data_pos == 0 or self.timepatch == "":
            raise ValueError("Wrong input detected.")

        data_length_dict = self.create_data_length_dict()
        # Make sure we read the exact number of lines we were asked to
        if self.is_binary:
            data_length = data_length_dict[self.timepatch] // 8
        else:
            data_length = data_length_dict[self.timepatch] // 4 + 2

        if num_of_items == -1:
            num_of_lines_to_read = -1
        else:
            num_of_lines_to_read = int(num_of_items * data_length)

        # Read file
        if self.is_binary:
            with open(self.filename, "rb") as f:
                f.seek(self.start_of_data_pos)
                arr: np.ndarray = np.fromfile(
                    f, dtype=f"u{data_length}", count=num_of_lines_to_read
                )
            return arr

        else:
            with open(self.filename, "rb") as f:
                f.seek(self.start_of_data_pos)
                arr = np.fromfile(
                    f, dtype="{}S".format(data_length), count=num_of_lines_to_read
                ).astype("{}U".format(data_length))
            return arr

    def create_inputs_dict(self) -> Dict[str, str]:
        """
        Create a dictionary for all input channels. Currently allows for three channels.
        'Empty' channels will not be checked.
        """
        dict_of_inputs = {}

        if self.input_start != "Empty":
            dict_of_inputs[self.input_start] = "110"

        if self.input_stop1 != "Empty":
            dict_of_inputs[self.input_stop1] = "001"

        if self.input_stop2 != "Empty":
            dict_of_inputs[self.input_stop2] = "010"

        assert len(dict_of_inputs) >= 1
        assert "Empty" not in list(dict_of_inputs.keys())

        # Calculate the number of channels
        self.num_of_channels = sum([1 for key in dict_of_inputs if "PMT" in key])

        return dict_of_inputs

    def compare_recorded_and_input_channels(self) -> None:
        """
        Raise error if user gave wrong amount of inputs
        """

        # If a user assumes an input exists, but it doesn't - raise an error
        recorded = self.list_of_recorded_data_channels.count(True)
        input_chans = len(self.dict_of_input_channels)
        if recorded < input_chans:
            raise UserWarning(
                f"Wrong number of user inputs ({input_chans}) "
                + f"compared to number of actual inputs "
                + f"({recorded}) to the multiscaler."
            )

        help_dict = {"001": 0, "010": 1, "110": 2}

        for key in self.dict_of_input_channels:
            if not self.list_of_recorded_data_channels[
                help_dict[self.dict_of_input_channels[key]]
            ]:
                raise UserWarning(
                    f'Wrong channel specification - the key "{key}" is on an empty channel'
                    f" (number {int(self.dict_of_input_channels[key], 2)})."
                )

    def __parse_extra_metadata(self, metadata):
        """
        Update self.lst_metadata with some additional information

        :param metadata: Data from the start of the lst file.
        """
        list_to_parse = [
            "fstchan",
            "holdafter",
            "periods",
            "rtpreset",
            "cycles",
            "sequences",
            "range",
            "sweepmode",
            "fdac",
        ]
        self.lst_metadata = {}
        for cur_str in list_to_parse:
            self.__parse_str(metadata, cur_str)

        self.lst_metadata["mirror_phase"] = str(self.mirror_phase)

    def __parse_str(self, metadata, str_to_parse):
        """
        Find str_to_parse in metadata and place the corresponding value in
        the dictionary self.lst_metadata
        """
        if self.is_binary:
            format_str: str = str_to_parse.encode("utf-8") + b"=(\w+)"
        else:
            format_str: str = str_to_parse + "=(\w+)"

        format_regex = re.compile(format_str)
        try:
            self.lst_metadata[str_to_parse] = str(
                re.search(format_regex, metadata).group(1)
            )
        except AttributeError:  # field is non-existent
            pass

    def __calc_actual_fill_fraction(self) -> float:
        """
        If we're using sweeps as lines then the true fill fraction is determined
        by the multiscaler's parameters, like hold_after and acquisiton delay.
        :return float: True fill fraction
        """
        if self.use_sweeps:
            fill_frac = self.data_range / (
                self.acq_delay + self.data_range + self.time_after
            )
            return fill_frac * 100  # in percent
        else:
            return -1.0



"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.nd_hist_generator.line_signal_validators.validation_tools import *
import pandas as pd
import numpy as np


class TestFrame(TestCase):
    """
    Tests for the validation functions
    """
    dict_of_data = dict(PMT1=pd.DataFrame([1, 10, 20, 30], columns=['abs_time']),
                        Lines=pd.DataFrame([0, 5, 10, 15, 20, 25, 30, 35], columns=['abs_time']))
    vlad = SignalValidator(dict_of_data)

    def test_last_event_only_pmt(self):
        dict_of_data = {'PMT1': pd.DataFrame([1, 5, 3], columns=['abs_time'])}
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 5)

    def test_last_event_two_pmts(self):
        dict_of_data = {'PMT1': pd.DataFrame([1, 5, 3], columns=['abs_time']),
                        'PMT2': pd.DataFrame([1, 2, 3], columns=['abs_time'])}
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 5)

    def test_last_event_with_frames(self):
        frame_data = pd.DataFrame([0, 100], columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Frames': frame_data,
                        'Lines': [1, 2, 3]}
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 200)

    def test_last_event_with_single_frame(self):
        frame_data = pd.DataFrame([100], columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Frames': frame_data,
                        'Lines': [1, 2, 3]}
        last = SignalValidator(dict_of_data)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 200)

    def test_last_event_with_lines_less_than_needed_single_frame(self):
        line_data = pd.DataFrame([0, 10, 20], columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        last = SignalValidator(dict_of_data, num_of_lines=5)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 50)

    def test_last_event_with_lines_less_than_needed_more_frames(self):
        line_data = pd.DataFrame([0, 10, 20, 30, 40], columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        last = SignalValidator(dict_of_data, num_of_lines=3)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 30)

    def test_last_event_with_lines_just_like_needed(self):
        line_data = pd.DataFrame([0, 10, 20], columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        last = SignalValidator(dict_of_data, num_of_lines=3)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 30)

    def test_last_event_with_lines_more_than_needed(self):
        line_data = pd.DataFrame(np.arange(0, 100, 10), columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        last = SignalValidator(dict_of_data, num_of_lines=3)._SignalValidator__calc_last_event_time()
        self.assertEqual(last, 90)

    def test_bins_bet_lines(self):
        line_freq = 10.
        binwidth = 0.01
        bidir = False
        bins = SignalValidator(dict_of_data=self.dict_of_data, line_freq=line_freq, binwidth=binwidth,
                               bidir=bidir)._SignalValidator__bins_bet_lines()
        self.assertEqual(10, bins)

    def test_bins_bet_lines_bidir(self):
        line_freq = 10.
        binwidth = 0.01
        bidir = True
        bins = SignalValidator(dict_of_data=self.dict_of_data, line_freq=line_freq, binwidth=binwidth,
                               bidir=bidir)._SignalValidator__bins_bet_lines()
        self.assertEqual(5, bins)

    def test_extrapolate_without_zero(self):
        sigval = SignalValidator(self.dict_of_data, line_freq=0.5, binwidth=1.,
                                 delay_between_frames=0.)
        line_point = self.dict_of_data['Lines'].iat[1, 0]
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta)
        returned_lines = sigval._SignalValidator__extrapolate_line_data(line_point=line_point)
        real_lines = np.arange(1, 15, step=2, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_extrapolate_from_zero(self):
        sigval = SignalValidator(self.dict_of_data, line_freq=1., binwidth=1.,
                                 delay_between_frames=0.)
        line_point = self.dict_of_data['Lines'].iat[2, 0]
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta)
        returned_lines = sigval._SignalValidator__extrapolate_line_data(line_point=line_point)
        real_lines = np.arange(0, 15, step=1, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_extrapolate_without_timepoint(self):
        sigval = SignalValidator(self.dict_of_data, line_freq=1., binwidth=1.,
                                 delay_between_frames=0.)
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta)
        returned_lines = sigval._SignalValidator__extrapolate_line_data()
        real_lines = np.arange(0, 15, step=1, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_extrapolate_without_timepoint_and_zero(self):
        sigval = SignalValidator(self.dict_of_data, line_freq=0.5, binwidth=1.,
                                 delay_between_frames=0.)
        sigval.last_event_time = 15
        sigval.line_delta = sigval._SignalValidator__bins_bet_lines()
        sigval.num_of_lines = int(sigval.last_event_time // sigval.line_delta) + 1
        returned_lines = sigval._SignalValidator__extrapolate_line_data()
        real_lines = np.arange(0, 15, step=2, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines.abs_time.tolist(), real_lines.tolist())

    def test_pipeline_single_frame_si(self):
        lines = [0, 5, 10, 15, 20, 25, 30, 70, 75]
        dict_of_data = dict(PMT1=pd.DataFrame([1, 10, 20, 30, 100], columns=['abs_time'], dtype=np.uint64),
                            Lines=pd.DataFrame(lines, columns=['abs_time'],
                                               dtype=np.uint64))
        binwidth = 1.
        line_freq = 0.2
        delay_bet_frames = 0.
        num_of_lines = 7
        sigval = SignalValidator(dict_of_data=dict_of_data, binwidth=binwidth,
                                 line_freq=line_freq, delay_between_frames=delay_bet_frames,
                                 num_of_lines=num_of_lines, image_soft=ImagingSoftware.SCANIMAGE.value)
        sigval.run()
        self.assertSequenceEqual(sigval.dict_of_data['Lines'].abs_time.tolist(), lines[:-2])
        self.assertSequenceEqual(sigval.dict_of_data['Frames'].abs_time.tolist(), [0])
        self.assertEqual(sigval.line_delta, 5)


    def test_pipeline_few_frames_si(self):
        lines = [0, 5, 10, 45, 50, 55, 90, 95, 100, 135, 140]
        dict_of_data = dict(PMT1=pd.DataFrame([1, 10, 20, 30, 100], columns=['abs_time'], dtype=np.uint64),
                            Lines=pd.DataFrame(lines, columns=['abs_time'], dtype=np.uint64))
        binwidth = 1.
        line_freq = 0.2
        delay_bet_frames = 0.
        num_of_lines = 3
        sigval = SignalValidator(dict_of_data=dict_of_data, binwidth=binwidth,
                                 line_freq=line_freq, delay_between_frames=delay_bet_frames,
                                 num_of_lines=num_of_lines, image_soft=ImagingSoftware.SCANIMAGE.value)
        sigval.run()
        self.assertSequenceEqual(sigval.dict_of_data['Lines'].abs_time.tolist(), lines[:-2])
        self.assertSequenceEqual(sigval.dict_of_data['Frames'].abs_time.tolist(), [0, 45, 90])
        self.assertEqual(sigval.line_delta, 5)

    def test_pipeline_single_frame_mscan(self):
        lines = [0, 5, 10, 15, 20, 25, 30, 35, 40]
        dict_of_data = dict(PMT1=pd.DataFrame([1, 10, 20, 30, 100], columns=['abs_time'], dtype=np.uint64),
                            Lines=pd.DataFrame(lines, columns=['abs_time'],
                                               dtype=np.uint64))
        binwidth = 1.
        line_freq = 0.2
        delay_bet_frames = 0.
        num_of_lines = 7
        sigval = SignalValidator(dict_of_data=dict_of_data, binwidth=binwidth,
                                 line_freq=line_freq, delay_between_frames=delay_bet_frames,
                                 num_of_lines=num_of_lines, image_soft=ImagingSoftware.MSCAN.value)
        sigval.run()
        self.assertSequenceEqual(sigval.dict_of_data['Lines'].abs_time.tolist(), lines[:-2])
        self.assertSequenceEqual(sigval.dict_of_data['Frames'].abs_time.tolist(), [0])
        self.assertEqual(sigval.line_delta, 5)

    def test_pipeline_few_frames_mscan(self):
        lines = [0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60]
        dict_of_data = dict(PMT1=pd.DataFrame([1, 10, 20, 30, 100, 200], columns=['abs_time'], dtype=np.uint64),
                            Lines=pd.DataFrame(lines, columns=['abs_time'], dtype=np.uint64))
        binwidth = 1.
        line_freq = 0.2
        delay_bet_frames = 0.
        num_of_lines = 3
        sigval = SignalValidator(dict_of_data=dict_of_data, binwidth=binwidth,
                                 line_freq=line_freq, delay_between_frames=delay_bet_frames,
                                 num_of_lines=num_of_lines, image_soft=ImagingSoftware.MSCAN.value)
        sigval.run()
        self.assertSequenceEqual(sigval.dict_of_data['Lines'].abs_time.tolist(), lines[:-1])
        self.assertSequenceEqual(sigval.dict_of_data['Frames'].abs_time.tolist(), [0, 15, 30, 45])
        self.assertEqual(sigval.line_delta, 5)



import unittest

from pysight.nd_hist_generator.allocation import *


class TestAllocation(unittest.TestCase):

    df_photons = pd.DataFrame(np.arange(10), columns=['abs_time'])
    dict_of_data = dict(Frames=pd.DataFrame(np.arange(0, 100, 10, dtype=np.uint64),
                                            columns=['abs_time']),
                        Lines=pd.DataFrame(np.arange(0, 100, 5, dtype=np.uint64),
                                           columns=['abs_time']))
    # allocat = Allocate(df_photons.copy(), dict_of_data.copy())

    def test_reindex_dict_of_data_standard_frames(self):
        allocat = Allocate(self.df_photons.copy(), self.dict_of_data.copy())
        allocat._Allocate__reindex_dict_of_data()
        true_frames = pd.Series(data=self.dict_of_data['Frames'].abs_time.values,
                                index=self.dict_of_data['Frames'].abs_time.values)
        np.testing.assert_equal(allocat.dict_of_data['Frames'].values,
                                true_frames.values)
        np.testing.assert_equal(allocat.dict_of_data['Frames'].index.values,
                                true_frames.index.values)

    def test_reindex_dict_of_data_standard_lines(self):
        allocat = Allocate(self.df_photons.copy(), self.dict_of_data.copy())
        allocat._Allocate__reindex_dict_of_data()
        lines = self.dict_of_data['Lines'].abs_time.values
        frames = np.repeat(self.dict_of_data['Frames'].abs_time.values.copy(), 2)
        np.testing.assert_equal(allocat.dict_of_data['Lines'].values,
                                lines)
        np.testing.assert_equal(allocat.dict_of_data['Lines'].index.values,
                                frames)

    def test_reindex_dict_of_data_lines_before_start_of_frames(self):
        frames = pd.DataFrame(np.arange(30, 100, 10, dtype=np.uint64),
                              columns=['abs_time'])
        lines = pd.DataFrame(np.arange(0, 100, 5, dtype=np.uint64),
                             columns=['abs_time'])
        dict_of_data = dict(Frames=frames, Lines=lines)
        allocat = Allocate(self.df_photons.copy(), dict_of_data)
        allocat._Allocate__reindex_dict_of_data()
        lines = lines.abs_time.values
        np.testing.assert_equal(allocat.dict_of_data['Lines'].values,
                                lines[6:])
        np.testing.assert_equal(allocat.dict_of_data['Lines'].index.values,
                                np.repeat(frames.abs_time.values.copy(), 2))

    def test_reindex_dict_of_data_lines_much_after_last_frame(self):
        frames = pd.DataFrame(np.arange(30, 100, 10, dtype=np.uint64),
                              columns=['abs_time'])
        lines = pd.DataFrame(np.arange(0, 150, 5, dtype=np.uint64),
                             columns=['abs_time'])
        dict_of_data = dict(Frames=frames, Lines=lines)
        allocat = Allocate(self.df_photons.copy(), dict_of_data)
        allocat._Allocate__reindex_dict_of_data()
        lines = lines.abs_time.values
        np.testing.assert_equal(allocat.dict_of_data['Lines'].values,
                                lines[6:])
        frames = np.repeat(frames.abs_time.values.copy(), 2)
        frames = np.concatenate((frames, np.repeat(np.array([90]), 10)))
        np.testing.assert_equal(allocat.dict_of_data['Lines'].index.values, frames)
